use std::thread;

use anyhow::Result;
use embedded_svc::{
    event_bus::EventBus,
    mqtt::client::{utils::ConnState, Client, Connection, MessageImpl, Publish, QoS},
};
use esp_idf_svc::{
    eventloop::{
        Background, EspBackgroundEventLoop, EspEventFetchData, EspEventLoop, EspEventPostData,
        EspSubscription, EspTypedEventDeserializer, EspTypedEventSerializer, EspTypedEventSource,
        User,
    },
    mqtt::client::{EspMqttClient, MqttClientConfiguration},
};
use esp_idf_sys::{c_types, EspError};
use log::*;
use serde::{Deserialize, Serialize};

pub static mut MQ_EVENTLOOP: Option<EspEventLoop<esp_idf_svc::eventloop::User<Background>>> = None;

pub struct MessageQueue {
    pub client: Option<EspMqttClient<ConnState<MessageImpl, EspError>>>,
    voltage_subscription: Option<EspSubscription<User<Background>>>,
}

impl MessageQueue {
    pub fn new() -> MessageQueue {
        return MessageQueue {
            client: None,
            voltage_subscription: None,
        };
    }

    pub fn init(&mut self) -> Result<()> {
        let conf = MqttClientConfiguration {
            client_id: Some("rust-esp32-std-demo"),
            crt_bundle_attach: Some(esp_idf_sys::esp_crt_bundle_attach),

            ..Default::default()
        };

        let (mut client, mut connection) =
            EspMqttClient::new_with_conn("mqtt://192.168.31.11:1883", &conf)?;

        info!("MQTT client started");
        thread::spawn(move || {
            info!("MQTT Listening for messages");

            while let Some(msg) = connection.next() {
                match msg {
                    Err(e) => info!("MQTT Message ERROR: {}", e),
                    Ok(msg) => info!("MQTT Message: {:?}", msg),
                }
            }

            info!("MQTT connection loop exit");
        });

        client
            .subscribe("esp32-c3-rust-wifi-demo", QoS::AtMostOnce)
            .map_err(|err| anyhow::anyhow!("subscribe message from queue failed. {}", err))?;

        info!("Subscribed to all topics (esp32-c3-rust-wifi-demo)");

        client
            .publish(
                "esp32-c3-rust-wifi-demo/ping",
                QoS::AtMostOnce,
                false,
                "Hello".as_bytes(),
            )
            .map_err(|err| anyhow::anyhow!("publish message to queue failed. {}", err))?;

        info!("Published a hello message to topic \"esp32-c3-rust-wifi-demo/ping\"");

        self.client = Some(client);
        anyhow::Ok(())
    }

    pub fn publish(&mut self, topic: &str, bytes: &[u8]) -> Result<u32> {
        self.client
            .as_mut()
            .unwrap()
            .publish(
                format!("{}/{}", "ups_0_2", topic).as_str(),
                QoS::AtMostOnce,
                false,
                bytes,
            )
            .map_err(|err| anyhow::anyhow!("publish message to queue was failed!. {}", err))
    }

    pub fn watch(mut self) -> anyhow::Result<EspSubscription<User<Background>>> {
        match EspBackgroundEventLoop::new(&Default::default()) {
            Ok(eventloop) => unsafe { MQ_EVENTLOOP = Some(eventloop) },
            Err(err) => anyhow::bail!("Init Event Loop failed. {:?}", err),
        }
        if let Some(eventloop) = unsafe { MQ_EVENTLOOP.as_mut() } {
            let subscription = eventloop
                .subscribe(move |dto: &MqDto| {
                    info!(
                        "publish data: {:?}",
                        unsafe {
                            String::from_utf8_unchecked(dto.message.to_vec())
                        }
                    );
                    if let Err(err) = self.publish(dto.topic, dto.message) {
                        warn!("Can not publish message to MQTT. {}", err);
                    }
                })
                .map_err(|err| anyhow::anyhow!(" Listening Event Loop Failed. {}", err))?;
            // self.voltage_subscription = Some(subscription);
            return anyhow::Ok(subscription);
        }
        anyhow::bail!("MQ_EVENTLOOP is Not Defined.")
    }
}

#[derive(Copy, Debug, Clone, Serialize, Deserialize)]
pub struct MqDto<'a, 'b> {
    pub topic: &'a str,
    pub message: &'b [u8],
}

impl EspTypedEventSource for MqDto<'_, '_> {
    fn source() -> *const c_types::c_char {
        b"MqDto\0".as_ptr() as *const _
    }
}

impl EspTypedEventSerializer<MqDto<'_, '_>> for MqDto<'_, '_> {
    fn serialize<R>(event: &MqDto, f: impl for<'a> FnOnce(&'a EspEventPostData) -> R) -> R {
        f(&unsafe { EspEventPostData::new(Self::source(), Self::event_id(), event) })
    }
}

impl<'b, 'c> EspTypedEventDeserializer<MqDto<'b, 'c>> for MqDto<'b, 'c> {
    fn deserialize<R>(
        data: &EspEventFetchData,
        f: &mut impl for<'a> FnMut(&'a MqDto<'b, 'c>) -> R,
    ) -> R {
        f(unsafe { data.as_payload() })
    }
}
